from copy import deepcopy
import math
import shutil
import statistics
import sys
from pathlib import Path
from typing import Any, Literal

import delu
import numpy as np
import rtdl_num_embeddings
import scipy
import torch
import torch.nn as nn
import torch.utils.tensorboard
from loguru import logger
from torch import Tensor
from tqdm import tqdm
from typing_extensions import NotRequired, TypedDict

if __name__ == '__main__':
    _cwd = Path.cwd()
    assert _cwd.joinpath(
        'ROOT_FILE'
    ).exists(), 'The script must be run from the root of the repository'
    sys.path.append(str(_cwd))
    del _cwd

import lib
import lib.data
import lib.deep
import lib.env
import lib.metrics
from lib import KWArgs, PartKey


def _get_first_input_scaling(
    backbone: lib.deep.MLP | lib.deep.ResNet,
) -> lib.deep.LinearEfficientEnsemble:
    if isinstance(backbone, lib.deep.MLP):
        return backbone.blocks[0][0]  # type: ignore[code]
    elif isinstance(backbone, lib.deep.ResNet):
        return backbone.blocks[0][1] if backbone.proj is None else backbone.proj  # type: ignore[code]
    else:
        raise RuntimeError(f'Unsupported backbone: {backbone}')


@torch.inference_mode()
def _init_scaling_by_sections(
    weight: Tensor,
    distribution: Literal['normal', 'random-signs'],
    init_sections: list[int],
) -> None:
    """Initialize the (typically, first) scaling in a special way.

    For a given efficient emsemble member, all weights within one section
    are initialized with the same value.
    Typically, one section corresponds to one feature.
    """
    assert weight.ndim == 2
    assert weight.shape[1] == sum(init_sections)

    if distribution == 'normal':
        init_fn_ = nn.init.normal_
    elif distribution == 'random-signs':
        init_fn_ = lib.deep.init_random_signs_
    else:
        raise ValueError(f'Unknown distribution: {distribution}')

    section_bounds = [0, *torch.tensor(init_sections).cumsum(0).tolist()]
    for i in range(len(init_sections)):
        w = torch.empty((len(weight), 1), dtype=weight.dtype, device=weight.device)
        init_fn_(w)
        weight[:, section_bounds[i] : section_bounds[i + 1]] = w

class Model(nn.Module):
    def __init__(
        self,
        *,
        n_num_features: int,
        cat_cardinalities: list[int],
        n_classes: None | int,
        bins: None | list[Tensor],
        backbone: dict,
        num_embeddings: None | dict = None,
        arch_type: Literal[
            # Active
            'vanilla', # Simple MLP
            'tabm',  # BatchEnsemble + separate heads + better initialization
            'tabm-mini',  # Minimal: * weight
            # BatchEnsemble
            'tabm-naive'
        ],
        k: None | int = None,
    ) -> None:
        # >>> Validate arguments.
        assert n_num_features >= 0
        assert n_num_features or cat_cardinalities
        if arch_type == 'vanilla':
            assert k is None
        else:
            assert k is not None
            assert k > 0

        super().__init__()

        # >>> Continuous (numerical) features
        scaling_init_sections = []

        if n_num_features == 0:
            assert bins is None
            self.num_module = None
            d_num = 0

        elif num_embeddings is None:
            assert bins is None
            self.num_module = None
            d_num = n_num_features
            scaling_init_sections.extend(1 for _ in range(n_num_features))

        else:
            if bins is None:
                self.num_module = lib.deep.make_module(
                    **num_embeddings, n_features=n_num_features
                )
            else:
                assert num_embeddings['type'].startswith('PiecewiseLinearEmbeddings')
                self.num_module = lib.deep.make_module(**num_embeddings, bins=bins)
            d_num = n_num_features * num_embeddings['d_embedding']
            scaling_init_sections.extend(
                num_embeddings['d_embedding'] for _ in range(n_num_features)
            )

        # >>> Categorical features
        self.cat_module = (
            lib.deep.OneHotEncoding0d(cat_cardinalities) if cat_cardinalities else None
        )
        scaling_init_sections.extend(cat_cardinalities)
        d_cat = sum(cat_cardinalities)

        # >>> Backbone
        d_flat = d_num + d_cat
        self.affine_ensemble = None
        self.backbone = lib.deep.make_module(d_in=d_flat, **backbone)

        if arch_type != 'vanilla':
            assert k is not None
            scaling_init = (
                'random-signs'
                if num_embeddings is None
                else 'normal'
            )

            if arch_type == 'tabm-mini':
                # The minimal possible efficient ensemble.
                self.affine_ensemble = lib.deep.ElementwiseAffineEnsemble(
                    k,
                    d_flat,
                    weight=True,
                    bias=False,
                    weight_init=(
                        'random-signs'
                        if num_embeddings is None
                        else 'normal'
                    ),
                )
                _init_scaling_by_sections(
                    self.affine_ensemble.weight,  # type: ignore[code]
                    scaling_init,
                    scaling_init_sections,
                )

            elif arch_type == 'tabm-naive':
                # The original BatchEnsemble.
                lib.deep.make_efficient_ensemble(
                    self.backbone,
                    k=k,
                    ensemble_scaling_in=True,
                    ensemble_scaling_out=True,
                    ensemble_bias=True,
                    scaling_init='random-signs',
                )
            elif arch_type == 'tabm':
                # Like BatchEnsemble, but all scalings, except for the first one,
                # are initialized with ones.
                lib.deep.make_efficient_ensemble(
                    self.backbone,
                    k=k,
                    ensemble_scaling_in=True,
                    ensemble_scaling_out=True,
                    ensemble_bias=True,
                    scaling_init='ones',
                )
                _init_scaling_by_sections(
                    _get_first_input_scaling(self.backbone).r,  # type: ignore[code]
                    scaling_init,
                    scaling_init_sections,
                )

            else:
                raise ValueError(f'Unknown arch_type: {arch_type}')

        # >>> Output
        d_block = backbone['d_block']
        d_out = 1 if n_classes is None else n_classes
        self.output = (
            nn.Linear(d_block, d_out)
            if arch_type == 'vanilla'
            else delu.nn.NLinear(k, d_block, d_out)  # type: ignore[code]
        )

        # >>>
        self.arch_type = arch_type
        self.k = k

    def forward(
        self, *, x_num: None | Tensor = None, x_cat: None | Tensor = None
    ) -> Tensor:
        x = []
        if x_num is not None:
            x.append(x_num if self.num_module is None else self.num_module(x_num))
        if x_cat is None:
            assert self.cat_module is None
        else:
            assert self.cat_module is not None
            x.append(self.cat_module(x_cat))
        x = torch.column_stack([x_.flatten(1, -1) for x_ in x])
        if x.dtype == torch.int64:
            x = x.float()

        if self.k is not None:
            x = x[:, None].expand(-1, self.k, -1)  # (B, D) -> (B, K, D)
            if self.affine_ensemble is not None:
                x = self.affine_ensemble(x)
        else:
            assert self.affine_ensemble is None

        x = self.backbone(x)
        x = self.output(x)
        if self.k is None:
            # Adjust the output shape for vanilla networks to make them compatible
            # with the rest of the script (loss, metrics, predictions, ...).
            # (B, D_OUT) -> (B, 1, D_OUT)
            x = x[:, None]
        return x

class Config(TypedDict):
    seed: int
    data: KWArgs
    bins: NotRequired[KWArgs]
    model: KWArgs
    head_selection: NotRequired[bool]
    optimizer: KWArgs
    n_lr_warmup_epochs: NotRequired[int]
    batch_size: int
    patience: int
    n_epochs: int
    gradient_clipping_norm: NotRequired[float]
    parameter_statistics: NotRequired[bool]
    amp: NotRequired[bool]  # Automatic mixed precision in bfloat16.
    #
    n_objects: int
    calculate_grad_sims: NotRequired[bool]
    calculate_grad_coeff_var: NotRequired[bool]


def main(
    config: Config | str | Path,
    output: None | str | Path = None,
    *,
    force: bool = False,
) -> None | lib.JSONDict:
    # >>> Start
    config, output = lib.check(config, output, config_type=Config)
    if not lib.start(output, force=force):
        return None

    lib.print_config(config)  # type: ignore[code]
    delu.random.seed(config['seed'])
    device = lib.get_device()
    report = lib.create_report(main, config)

    # >>> Data
    dataset = lib.data.build_dataset(**config['data'])
    if dataset.task.is_regression:
        dataset.data['y'], regression_label_stats = lib.data.standardize_labels(
            dataset.data['y']
        )
    else:
        regression_label_stats = None

    # Convert binary features to categorical features.
    if dataset.n_bin_features > 0:
        x_bin = dataset.data.pop('x_bin')
        # Remove binary features with just one unique value in the training set.
        # This must be done, otherwise, the script will fail on one specific dataset
        # from the "why" benchmark.
        n_bin_features = x_bin['train'].shape[1]
        good_bin_idx = [
            i for i in range(n_bin_features) if len(np.unique(x_bin['train'][:, i])) > 1
        ]
        if len(good_bin_idx) < n_bin_features:
            x_bin = {k: v[:, good_bin_idx] for k, v in x_bin.items()}

        if dataset.n_cat_features == 0:
            dataset.data['x_cat'] = {
                part: np.zeros((dataset.size(part), 0), dtype=np.int64)
                for part in x_bin
            }
        for part in x_bin:
            dataset.data['x_cat'][part] = np.column_stack(
                [dataset.data['x_cat'][part], x_bin[part].astype(np.int64)]
            )
        del x_bin
    dataset = dataset.to_torch(device)
    Y_train = dataset.data['y']['train'].to(
        torch.long if dataset.task.is_classification else torch.float
    )

    # >>> Model
    if 'bins' in config:
        # Compute the bins for PiecewiseLinearEncoding and PiecewiseLinearEmbeddings.
        compute_bins_kwargs = (
            {
                'y': Y_train.to(
                    torch.long if dataset.task.is_classification else torch.float
                ),
                'regression': dataset.task.is_regression,
                'verbose': True,
            }
            if 'tree_kwargs' in config['bins']
            else {}
        )
        bin_edges = rtdl_num_embeddings.compute_bins(
            dataset['x_num']['train'], **config['bins'], **compute_bins_kwargs
        )
        logger.info(f'Bin counts: {[len(x) - 1 for x in bin_edges]}')
    else:
        bin_edges = None
    model = Model(
        n_num_features=dataset.n_num_features,
        cat_cardinalities=dataset.compute_cat_cardinalities(),
        n_classes=dataset.task.try_compute_n_classes(),
        **config['model'],
        bins=bin_edges,
    )
    report['n_parameters'] = lib.deep.get_n_parameters(model)
    logger.info(f'n_parameters = {report["n_parameters"]}')
    report['prediction_type'] = 'labels' if dataset.task.is_regression else 'probs'
    model.to(device)
    if torch.cuda.device_count() > 1:
        model = nn.DataParallel(model)

    # >>> Training
    step = 0
    batch_size = config['batch_size']
    report['epoch_size'] = epoch_size = math.ceil(dataset.size('train') / batch_size)
    eval_batch_size = 32768
    chunk_size = None

    optimizer = lib.deep.make_optimizer(
        **config['optimizer'], params=lib.deep.make_parameter_groups(model)
    )
    gradient_clipping_norm = config.get('gradient_clipping_norm')
    n_lr_warmup_epochs = config.get('n_lr_warmup_epochs')
    if n_lr_warmup_epochs is not None and n_lr_warmup_epochs > 0:
        n_warmup_steps = n_lr_warmup_epochs * epoch_size
        n_warmup_steps = max(1, math.trunc(n_warmup_steps / epoch_size)) * epoch_size
        logger.info(f'{n_warmup_steps=}')
        lr_scheduler = torch.optim.lr_scheduler.LinearLR(
            optimizer, start_factor=0.01, total_iters=n_warmup_steps
        )
    else:
        lr_scheduler = None
    _loss_fn = (
        nn.functional.mse_loss
        if dataset.task.is_regression
        else nn.functional.cross_entropy
    )

    def loss_fn(y_pred: Tensor, y_true: Tensor) -> Tensor:
        return _loss_fn(y_pred.flatten(0, 1), y_true.repeat_interleave(y_pred.shape[1]))

    # The following generator is used only for creating training batches,
    # so the random seed fully determines the sequence of training objects.
    batch_generator = torch.Generator(device).manual_seed(config['seed'])
    timer = delu.tools.Timer()
    early_stopping = delu.tools.EarlyStopping(config['patience'], mode='max')
    parameter_statistics = config.get('parameter_statistics', config['seed'] == 1)
    training_log = []
    writer = torch.utils.tensorboard.SummaryWriter(output)  # type: ignore[code]
    amp_enabled = (
        config.get('amp', False)
        and device.type == 'cuda'
        and torch.cuda.is_bf16_supported()
    )
    logger.info(f'AMP enabled: {amp_enabled}')

    @torch.autocast(  # type: ignore[code]
        device.type, enabled=amp_enabled, dtype=torch.bfloat16 if amp_enabled else None
    )
    def apply_model(part: PartKey, idx: Tensor) -> Tensor:
        return (
            model(
                **{
                    key: dataset.data[key][part][idx]
                    for key in ['x_num', 'x_bin', 'x_cat']
                    if key in dataset
                }
            )
            .squeeze(-1)
            .float()
        )

    @torch.inference_mode()
    def evaluate(
        parts: list[PartKey], eval_batch_size: int
    ) -> tuple[
        dict[PartKey, Any], dict[PartKey, np.ndarray], dict[PartKey, np.ndarray], int
    ]:
        model.eval()
        head_predictions: dict[PartKey, np.ndarray] = {}
        for part in parts:
            while eval_batch_size:
                try:
                    head_predictions[part] = (
                        torch.cat(
                            [
                                apply_model(part, idx)
                                for idx in torch.arange(
                                    len(dataset.data['y'][part]),
                                    device=device,
                                ).split(eval_batch_size)
                            ]
                        )
                        .cpu()
                        .numpy()
                    )
                except RuntimeError as err:
                    if not lib.is_oom_exception(err):
                        raise
                    eval_batch_size //= 2
                    logger.warning(f'eval_batch_size = {eval_batch_size}')
                else:
                    break
            if not eval_batch_size:
                RuntimeError('Not enough memory even for eval_batch_size=1')
        if dataset.task.is_regression:
            assert regression_label_stats is not None
            head_predictions = {
                k: v * regression_label_stats.std + regression_label_stats.mean
                for k, v in head_predictions.items()
            }
        else:
            head_predictions = {
                k: scipy.special.softmax(v, axis=-1)
                for k, v in head_predictions.items()
            }
            if dataset.task.is_binclass:
                head_predictions = {k: v[..., 1] for k, v in head_predictions.items()}

        predictions = {k: v.mean(1) for k, v in head_predictions.items()}
        metrics = (
            dataset.task.calculate_metrics(predictions, report['prediction_type'])
            if lib.are_valid_predictions(predictions)
            else {x: {'score': -999999.0} for x in predictions}
        )
        for k in parts:
            if dataset.task.is_regression:
                assert regression_label_stats is not None
                metrics[k]['loss'] = (
                    metrics[k]['rmse'] / regression_label_stats.std
                ) ** 2
            else:
                metrics[k]['loss'] = metrics[k]['cross-entropy']

        return metrics, predictions, head_predictions, eval_batch_size

    def save_checkpoint() -> None:
        lib.dump_checkpoint(
            output,
            {
                'step': step,
                'model': model.state_dict(),
                'optimizer': optimizer.state_dict(),
                'batch_generator': batch_generator.get_state(),
                'random_state': delu.random.get_state(),
                'early_stopping': early_stopping,
                'report': report,
                'timer': timer,
                'training_log': training_log,
            }
            | (
                {}
                if lr_scheduler is None
                else {'lr_scheduler': lr_scheduler.state_dict()}
            ),
        )
        lib.dump_report(output, report)
        lib.backup_output(output)

    k_model = config['model'].get('k', 1)
    # we choose object indices that will be used to calculate grad statistics
    object_indices = torch.randperm(
        dataset.size('train'), generator=torch.Generator().manual_seed(config['seed'])
    )[: config['n_objects']].tolist()
    report['indices'] = object_indices
    parameter_names = sorted(
        [name for name, _ in model.named_parameters() if 'output.' not in name]
    )

    # for cosine sim between grads of heads
    calculate_grad_sims = config.get('calculate_grad_sims', True)
    MAX_OBJECTS_FOR_GRAD_SIMS = 1000
    n_objects_for_grad_sims = min(config['n_objects'], MAX_OBJECTS_FOR_GRAD_SIMS)
    grad_similarities: dict[str, list] = {name: [] for name in parameter_names}
    grad_similarities['model'] = []
    grad_similarities['backbone'] = []

    n_layers = config['model']['backbone']['n_blocks']
    additional_keys = ['backbone'] + [
        f'backbone.blocks.{lix}' for lix in range(n_layers)
    ]
    for key in additional_keys:
        grad_similarities[key] = []

    # for coeff of variance of grad
    calculate_grad_coeff_var = config.get('calculate_grad_coeff_var', True)
    grad_norms_vars = {
        name: {k: [] for k in ['norm', 'std']} for name in grad_similarities.keys()
    }

    # for some scalar statistics
    grad_scalars_stats = deepcopy(grad_norms_vars)

    # here, we store gradients for each object.
    grads_for_objects = {
        name: torch.zeros(
            config['n_objects'], p.numel(), device='cpu', dtype=torch.float32
        )
        for name, p in model.named_parameters()
        if name in grad_norms_vars.keys()
    }

    grads_for_objects['model'] = torch.zeros(
        config['n_objects'],
        sum([p.shape[1] for p in grads_for_objects.values()]),
        device='cpu',
        dtype=torch.float32,
    )

    for key in additional_keys:
        grads_for_objects[key] = torch.zeros(
            config['n_objects'],
            sum(
                [
                    p.shape[1]
                    for name, p in grads_for_objects.items()
                    if name.startswith(key)
                ]
            ),
            device='cpu',
            dtype=torch.float32,
        )
    ###
    print(parameter_names)
    print()
    timer.run()
    while config['n_epochs'] == -1 or step // epoch_size < config['n_epochs']:
        print(f'[...] {lib.try_get_relative_path(output)} | {timer}')

        # >>> Compute similarities between the gradients produced by all heads.

        model.train()

        def calculate_gradient_similarities():
            epoch_placeholder = {
                k: np.zeros((k_model, k_model), dtype=np.float32)
                for k in grad_similarities.keys()
            }
            for i in tqdm(range(n_objects_for_grad_sims), ncols=0):
                index = object_indices[i]
                grads = {}
                for head_index in range(k_model):
                    optimizer.zero_grad()
                    _loss_fn(
                        apply_model('train', [index])[:, head_index],  # type: ignore[code]
                        Y_train[[index]],
                    ).backward()
                    for name in parameter_names:
                        parameter = model.get_parameter(name)
                        assert parameter.grad is not None
                        grads.setdefault(name, []).append(
                            parameter.grad.detach().flatten()
                        )
                optimizer.zero_grad()
                grads = {k: torch.stack(v) for k, v in grads.items()}
                grads.setdefault(
                    'model', torch.cat([grads[name] for name in parameter_names], -1)
                )
                for key in additional_keys:
                    grads.setdefault(
                        key,
                        torch.cat(
                            [
                                grads[name]
                                for name in parameter_names
                                if name.startswith(key)
                            ],
                            -1,
                        ),
                    )

                for name in grads.keys():
                    ng = nn.functional.normalize(grads[name], dim=-1)
                    epoch_placeholder[name] += ng.mm(ng.t()).cpu().numpy()
            for name in epoch_placeholder.keys():
                grad_similarities[name].append(
                    epoch_placeholder[name] / n_objects_for_grad_sims
                )

        if calculate_grad_sims:
            calculate_gradient_similarities()

        # >>> Compute coefficient of variance for gradients.
        # Also some scalar statistics.
        def compute_coeff_of_var(
            grad_norms_vars: dict[str, dict[str, list]],
            grad_scalars_stats: dict[str, dict[str, list]],
            head_index: int | None = None,
        ):
            model.train()
            # in this loop, we collect grads for each object
            for i, index in tqdm(enumerate(object_indices)):
                grads = {}
                optimizer.zero_grad()
                if head_index is None:
                    loss_fn(
                        apply_model('train', [index]),  # type: ignore[code]
                        Y_train[[index]],
                    ).backward()
                else:
                    _loss_fn(
                        apply_model('train', [index])[:, head_index],  # type: ignore[code]
                        Y_train[[index]],
                    ).backward()

                for name in parameter_names:
                    parameter = model.get_parameter(name)
                    assert parameter.grad is not None
                    grads[name] = parameter.grad.detach().flatten().cpu()

                optimizer.zero_grad()
                grads['model'] = torch.cat(
                    [grads[name] for name in parameter_names], -1
                )
                for key in additional_keys:
                    grads[key] = torch.cat(
                        [
                            grads[name]
                            for name in parameter_names
                            if name.startswith(key)
                        ],
                        -1,
                    )
                for name in grads.keys():
                    grads_for_objects[name][i, :] = grads[name]

            # in this loop, we calculate statistics
            for name, cur_g in grads_for_objects.items():
                cur_g = cur_g.to(device)
                norm_g = cur_g.norm(dim=1)
                var_g = (
                    norm_g.square().mean() - cur_g.mean(dim=0).norm().square()
                )  # E[|g|^2] - |E[g]|^2
                grad_norms_vars[name]['norm'].append(norm_g.mean().cpu().item())
                grad_norms_vars[name]['std'].append(var_g.sqrt().cpu().item())

                cur_g.abs_()
                mean_g = cur_g.mean(dim=0).mean()
                std_g = cur_g.std(dim=0).mean()
                grad_scalars_stats[name]['norm'].append(mean_g.cpu().item())
                grad_scalars_stats[name]['std'].append(std_g.cpu().item())

        if calculate_grad_coeff_var:
            compute_coeff_of_var(grad_norms_vars, grad_scalars_stats)
            optimizer.zero_grad()

        # >> Training
        model.train()
        epoch_losses = []
        for batch_idx in tqdm(
            torch.randperm(
                len(dataset.data['y']['train']),
                generator=batch_generator,
                device=device,
            ).split(batch_size),
            desc=f'Epoch {step // epoch_size} Step {step}',
        ):
            loss, new_chunk_size = lib.deep.zero_grad_forward_backward(
                optimizer,
                lambda idx: loss_fn(apply_model('train', idx), Y_train[idx]),
                batch_idx,
                chunk_size or batch_size,
            )

            if parameter_statistics and (
                step % epoch_size == 0  # The first batch of the epoch.
                or step // epoch_size == 0  # The first epoch.
            ):
                for k, v in lib.deep.compute_parameter_stats(model).items():
                    writer.add_scalars(k, v, step, timer.elapsed())
                    del k, v

            if gradient_clipping_norm is not None:
                nn.utils.clip_grad.clip_grad_norm_(
                    model.parameters(), gradient_clipping_norm
                )
            optimizer.step()

            if lr_scheduler is not None:
                lr_scheduler.step()
            step += 1
            epoch_losses.append(loss.detach())
            if new_chunk_size and new_chunk_size < (chunk_size or batch_size):
                chunk_size = new_chunk_size
                logger.warning(f'chunk_size = {chunk_size}')

        epoch_losses = torch.stack(epoch_losses).tolist()
        mean_loss = statistics.mean(epoch_losses)
        metrics, predictions, _, eval_batch_size = evaluate(
            ['val', 'test'], eval_batch_size
        )

        training_log.append(
            {'epoch-losses': epoch_losses, 'metrics': metrics, 'time': timer.elapsed()}
        )
        lib.print_metrics(mean_loss, metrics)
        writer.add_scalars('loss', {'train': mean_loss}, step, timer.elapsed())
        for part in metrics:
            writer.add_scalars(
                'score', {part: metrics[part]['score']}, step, timer.elapsed()
            )

        if (
            'metrics' not in report
            or metrics['val']['score'] > report['metrics']['val']['score']
        ):
            print('🌸 New best epoch! 🌸')
            report['best_step'] = step
            report['metrics'] = metrics
            save_checkpoint()
            lib.dump_predictions(output, predictions)

        early_stopping.update(metrics['val']['score'])
        if early_stopping.should_stop() or not lib.are_valid_predictions(predictions):
            break

        print()
    report['time'] = str(timer)
    
    # >>> save cosine sims
    if calculate_grad_sims:
        grad_similarities = {
            k: [np.stack(x) for x in v] for k, v in grad_similarities.items()
        }
        np.save(
            output / 'grad_similarities.npy',
            {k: np.stack(v) for k, v in grad_similarities.items()},  # type: ignore[code]
        )

    # >> save grad norms
    if calculate_grad_coeff_var:
        np.save(output / 'grad_norms_vars.npy', grad_norms_vars)  # type: ignore[code]
        np.save(output / 'grad_scalars.npy', grad_scalars_stats)  # type: ignore[code]
    # >>>
    if lib.get_checkpoint_path(output).exists():
        model.load_state_dict(lib.load_checkpoint(output)['model'])
    report['metrics'], predictions, head_predictions, _ = evaluate(
        ['train', 'val', 'test'], eval_batch_size
    )
    report['chunk_size'] = chunk_size
    report['eval_batch_size'] = eval_batch_size
    lib.dump_predictions(output, predictions)
    lib.dump_summary(output, lib.summarize(report))
    save_checkpoint()

    lib.finish(output, report)
    return report


if __name__ == '__main__':
    lib.configure_libraries()
    lib.run(main)
